#!/usr/bin/env perl
# includeres: include resources in PostScript file
#
# Copyright (C) Angus J. C. Duggan 1991-1995
# See file LICENSE for details.

use strict;
use warnings;

# Resource extensions
my %extn = ("font" => ".pfa", "file" => ".ps", "procset" => ".ps",
            "pattern" => ".pat", "form" => ".frm", "encoding" => ".enc");
# Resource types
my %type = ("%%BeginFile:" => "file", "%%BeginProcSet:" => "procset",
            "%%BeginFont:" => "font");

sub filename {			# make filename for resource in @_
  my($name);
  foreach (@_) {		# sanitise name
    s/[!()\$\#*&\\\|\`\'\"\~\{\}\[\]\<\>\?]//g;
    $name .= $_;
  }
  $name =~ s@.*/@@;		# drop directories
  die "Filename not found for resource ", join(" ", @_), "\n"
    if $name =~ /^$/;
  $name;
}

while (<>) {
  if (/^%%IncludeResource:/ || /^%%IncludeFont:/ || /^%%IncludeProcSet:/) {
    my ($comment, @res) = split(/\s+/);
    my $type = defined($type{$comment}) ? $type{$comment} : shift(@res);
    my $name = filename(@res);
    if (open(RES, $name) || open(RES, "$name$extn{$type}")) {
      print do { local $/; <RES> };
      close(RES);
    } else {
      print "%%IncludeResource: ", join(" ", $type, @res), "\n";
      print STDERR "Resource $name not found\n";
    }
  } else {
    print $_;
  }
}
